
namespace GossamerVR.MorphFilterPlus
{
    using System;
    using System.Text;
    using System.Collections;
    using System.Collections.Generic;
    using UnityEngine;

    public class Filter
    {
        /// <summary>
        /// A Vam-friendly enum, essentially, since
        /// VAM won't let us use enums in plugins.
        /// </summary>
        private class ParseMode
        {
            public const int None = 0;
            public const int Pending = 1;
            public const int Prefix = 2;
            public const int Antiprefix = 3;
            public const int Exclude = 4;
        }


        /// <summary>
        /// The content of the filter without some special characters;
        /// used primarily to determiine an exact match where applicable.
        /// </summary>
        public string Content;

        /// <summary>
        /// true if the filter was not created with an empty string.
        /// </summary>
        public bool IsValid;

        /// <summary>
        /// true if the filter must precisely match the content.
        /// </summary>
        public bool ExactMatch;

        /// <summary>
        /// strings that the morph name must contain
        /// </summary>
        public List<string> Contains;

        /// <summary>
        /// strings that the morph name must NOT contain
        /// </summary>
        public List<string> Without;

        /// <summary>
        /// strings that the morph name must start with
        /// </summary>
        public List<string> StartsWith;

        /// <summary>
        /// strings that the morph name must NOT start with
        /// </summary>
        public List<string> DoesNotStartWith;

        // when we create the filter, we pass the raw string and break it apart
        // into its separate constraints. A filter can include multiple requirements,
        // so each constraint is created as a list. We want the creation of the filter
        // to be the heaviest part, and the comparison to be the lightest,
        // since we compare it to many morphs at a time.

        public Filter(string filterdata)
        {            
            this.IsValid = !string.IsNullOrEmpty(filterdata);
            this.Contains = new List<string>();
            this.Without = new List<string>();
            this.DoesNotStartWith = new List<string>();
            this.StartsWith = new List<string>();
            if (this.IsValid)
            {                
                this.ExactMatch = (filterdata[0] == '=');
                this.Content = (this.ExactMatch ? filterdata.Substring(1) : filterdata);
                
                // if an exact match is required, we don't need any other constraints.
                // it'll either match the content or it won't.
                if (!this.ExactMatch)
                {
                    this.BuildConstraints(filterdata);
                }
            }
        }

        private void BuildConstraints(string filterdata)
        {
            StringBuilder stringBuilder = new StringBuilder();

            // constraints are separated by the plus sign so users can provide multiple requirements.
            bool capture = false;
            string[] constraintGroups = filterdata.Split(new char[] { '+' }, StringSplitOptions.None);
            foreach (string constraintGroup in constraintGroups)
            {
                int prevMode = ParseMode.None;
                int mode = ParseMode.None;
                
                // we read the constraint element character by character to determine the constraints.
                for (int i = 0; i < constraintGroup.Length; i++)
                {
                    char c = constraintGroup[i];
                    if (c == '-' || c=='!')
                    {
                        bool bang = c == '!';
                        // we want a bang '!' or double-dash '--' to mean "without" and a dash-dollar '-$' or bang dollar '!$'
                        // to mean 'Doesn't start with' so we want to jump ahead to see if either of those characters are next.
                        i++;
                        if (i >= constraintGroup.Length)
                        {
                            // if there is no next character then it doesn't matter, we're done.
                            break;
                        }

                        // if we find either of those valid characters, then we set our detected
                        // parse mode accordingly.
                        c = constraintGroup[i];                        
                        if (c == '$')                        
                        {                            
                            prevMode = mode;
                            mode = ParseMode.Antiprefix;
                            capture = true;
                        }
                        else if (!bang && c == '-')
                        {
                            prevMode = mode;
                            mode = ParseMode.Exclude;
                            capture = true;
                        }
                        else if (bang)
                        {
                            // if the bang isn't followed by a special character,
                            // we back up so the next character is written as normal.
                            prevMode = mode;
                            mode = ParseMode.Exclude;
                            capture = true;
                            i--;
                        }
                    }
                    else if (c == '$')
                    {
                        // without the dash, a dollar sign means it must start with.
                        prevMode = mode;
                        mode = ParseMode.Prefix;
                        capture = true;
                    }
                    else
                    {
                        // without any special character, we are building the string
                        // that is applied as the constraint.
                        stringBuilder.Append(c);
                    }
                    if (capture)
                    {
                        // if we detected a change in parse mode, we commit the 
                        // string we were building in the previous mode to the
                        // appropriate constraint, and then start again.
                        string constraint = stringBuilder.ToString();
                        stringBuilder = new StringBuilder();
                        switch (prevMode)
                        {                            
                            case ParseMode.Prefix:
                                this.StartsWith.Add(constraint);
                                break;
                            case ParseMode.Antiprefix:
                                this.DoesNotStartWith.Add(constraint);
                                break;
                            case ParseMode.Exclude:
                                this.Without.Add(constraint);
                                break;
                            default:
                                this.Contains.Add(constraint);
                                break;
                        }
                        prevMode = ParseMode.Pending;
                        capture = false;
                    }
                }

                // once we've run to the end of the constraint group,
                // we are probably still building a string, because people aren't going 
                // to end their filter with special characters for no reason.
                // so we commit that string based on the current mode.
                if (stringBuilder.Length > 0)
                {
                    string constraint = stringBuilder.ToString();
                    stringBuilder = new StringBuilder();
                    switch (mode)
                    {                        
                        case ParseMode.Prefix:
                            this.StartsWith.Add(constraint);
                            break;
                        case ParseMode.Antiprefix:
                            this.DoesNotStartWith.Add(constraint);
                            break;
                        case ParseMode.Exclude:
                            this.Without.Add(constraint);
                            break;
                        default:
                            this.Contains.Add(constraint);
                            break;
                    }
                }
            }            
        }
        

        /// <summary>
        /// Checks a string to see that it matches our constraints.
        /// This is used on morph names during the ApplyFilter process. 
        /// Most of the heavy lifting was done during the creation of the filter,
        /// so this should be relatively fast as far as string checks go.
        /// </summary>        
        public bool CompareString(string name)
        {
            // while normally I'm not a fan of early returns,
            // we want to bail on this filter as soon as possible for performance.
            // to prevent a ton of nesting we will return as soon as a constraint is failed
            // (or immediately in the case of an exact match fail or success)

            if (this.ExactMatch)
            {
                return name == this.Content;
            }

            foreach (string constraint in this.StartsWith)
            {
                if (!name.StartsWith(constraint))
                {
                    return false;
                }
            }

            foreach (string constraint in this.DoesNotStartWith)
            {
                if (name.StartsWith(constraint))
                {
                    return false;
                }
            }


            foreach (string constraint in this.Without)
            {
                if (name.Contains(constraint))
                {                   
                    return false;
                }
            }

            foreach (string constraint in this.Contains)
            {
                if (!name.Contains(constraint))
                {            
                    return false;
                }
            }


            return true;
        }

      

    }
}