
/**
 *  I originally created this to be a small, single script solution based on my library changes.
 *  however in the DLL i was free to modify a lot of things. Here, forced to work with only public 
 *  elements, and limited ui display interception, it's grown a little bit complicated.
 *  At some point I'll probably break this into multiple components or otherwise encapsulate concerns.
 *  in the meantime here's a big script.
 *  
 *  if it's any consolation it's 90% comments.
 * */


namespace GossamerVR.MorphFilterPlus
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Text.RegularExpressions;
    using UnityEngine;
    using UnityEngine.UI;
    using UnityEngine.Events;
    using UnityEngine.EventSystems;

    using SimpleJSON;


    /// <summary>
    /// Upgrades person morph search bars to have AND, NOT, OR, STARTSWITH and other filter controls.
    /// designed to be a session plugin
    /// </summary>
    public class MorphFilterPlus : MVRScript
    {
        // strings are dangerous. It's easy to misspell a string, and then 
        // you're spending hours trying to figure out what went wrong.
        // so we put our strings all in one place, defined in advance, up here.
        // By making them fields and constants instead of hand-written strings, the compiler
        // will alert us if we misspell anything when we use them, AND we
        // only have to correct or change the string in a single place.
        ////////////////////// STRINGS ///////////////////////////////
        #region strings

        // all transform paths are relative to an atom transform.

        /// <summary>
        /// the transform path to the child with an AtomUI component, 
        /// for which we will be modifying the SaveAppearance button, 
        /// so that we can respond before the save.
        /// </summary>
        private const string k_personUIPath = "UI/UIPlaceHolderModel/UIModel";
        private const string k_personGeomPath = "rescale2/geometry";
        private const string k_presetControlPath = k_personGeomPath + "/AppearancePresets";


        private const string k_personUIContentPath = k_personUIPath + "/Canvas/Panel/Content";        
        // there is a separate morph ui instance for both male and female pages.
        // they also change their name once they first populate, so we have to account for multiple names,
        // as we won't know if it's been populated when we do a search.
        private const string k_femaleMorphUIPath = k_personUIContentPath + "/Female Morphs";
        private const string k_femaleMorphUIPathAlt = k_personUIContentPath + "/TabMorphsFemale";
        private const string k_maleMorphUIPath = k_personUIContentPath + "/Male Morphs";
        private const string k_maleMorphUIPathAlt = k_personUIContentPath + "/TabMorphsMale";

        // these are the elements we might be looking for in the morph ui paths
        private const string k_femaleMorphUIControl = "MorphAutoUIFemale";
        private const string k_maleMorphUIControl = "MorphAutoUIMale";
        private const string k_femaleCategoryPopup = "MorphAutoUIFemaleCategoryPopup";
        private const string k_maleCategoryPopup = "MorphAutoUIMaleCategoryPopup";

        private const string k_filterInputField = "FilterInputField";
        private const string k_onlyFavToggle = "OnlyFavsToggle";
        private const string k_onlyLatestToggle = "OnlyLatestToggle";
        private const string k_onlyActiveToggle = "OnlyActiveToggle";
        private const string k_showTypePopup = "ShowPopup";
        


        /// <summary>
        /// path that we parent our appearance listener to. Adding a child to
        /// be considered 'part of the appearance' is the only way i've found so far
        /// to actually listen for the save of the appearance and react. 
        /// </summary>        
        private const string k_listenerParentPath = k_personGeomPath + "/FemaleCharacters";


        /// <summary>
        /// the array of morph banks collectively storing all the character morphs.
        /// we store them as an array so we can foreach() them easily. Their order
        /// is important because we'll use the MorphBankPath enum as indices.
        /// </summary>
        private static readonly string[] morphBankPaths = new[]
        {
            k_personGeomPath+ "/MorphBanks/FemaleCharacterMorphBank(Clone)",
            k_personGeomPath+ "/MorphBanks/FemaleGenMorphBank(Clone)",
            k_personGeomPath+ "/MorphBanks/MaleCharacterMorphBank(Clone)",
            k_personGeomPath+ "/MorphBanks/MaleGenMorphBank(Clone)",
            k_personGeomPath+ "/MorphBanks/MaleGen2MorphBank(Clone)",
        };

        public static class AtomType
        {
            public const string Person = "Person";
            public const string SubScene = "SubScene";
        }

        // while most of the time our bank paths are iterated over,
        // we are going to want to be able to access them specifically sometimes.
        // Rather than memorizing what each int index means, we use constants.
        public static class BankIndex
        {
            public const int FemaleChar = 0;
            public const int FemaleGen = 1;
            public const int MaleChar = 2;
            public const int MaleGen = 3;
            public const int MaleGen2 = 4;
        }
        // this is where I would be using an enum.
        //
        //                               IF I HAD ONE


        #endregion strings
        //////////////////////////////////////////////////////////////

        /// <summary>
        /// to let people know it's working we'll turn the input fields blue.
        /// </summary>
        private static readonly Color mintyFreshColor = new Color(0.85f, 0.95f, 1f, 1f);


        /// <summary>
        /// The UIDs of morphs VAM has already marked as invisible before we start.
        /// We store these to ensure we never set them visible.
        /// </summary>        
        public static HashSet<string> IgnoredMorphs = null;


        /// <summary>        
        /// when true, the filter will behave as normal: applying every time
        /// the input field changes. When it's false, the filter only gets applied
        /// when the user clicks away from the field or presses enter.
        /// We show this in the plugin's UI
        /// </summary>
        private JSONStorableBool liveUpdateBoolJSON;

        /// <summary>
        /// when true, spaces will be converted to '+' symbol as a QOL option.
        /// </summary>
        private JSONStorableBool convertSpacesBoolJSON;

        /// <summary>
        /// allows the script to log more.
        /// </summary>
        private JSONStorableBool debugBoolJSON;
        private bool isDebugMode => this.debugBoolJSON != null && this.debugBoolJSON.val;

        /// <summary>
        /// the number of characters in the input field required before 
        /// the filter will begin to apply if live update is true.
        /// </summary>
        private JSONStorableFloat minCharacterFloatJSON;

        /// <summary>
        /// A table of the atoms we have "upgraded" and their upgrades,
        /// which store references to modifications.
        /// </summary>      
        private Dictionary<Atom, AtomUpgrade> upgradeLookup = new Dictionary<Atom, AtomUpgrade>();

        /// <summary>
        /// ongoing coroutine tasks to move through in update
        /// because I prefer having direct control instead of using StartCoroutine
        /// </summary>
        private List<IEnumerator> tasks = new List<IEnumerator>();

        /// <summary>
        /// When this string is populated, it means something went wrong.
        /// it will be logged and will tell the plugin to shut down.
        /// it allows us to gracefully disable when there's an error,
        /// and not spam the user
        /// </summary>
        private string generalError = "";

        private bool hasError => this.everythingIsTerrible || !string.IsNullOrEmpty(generalError);

        /// <summary>
        /// This flag is set when we self-disable due to critical errors,        
        /// it allows us to sleep, essentially, until reset.
        /// </summary>
        private bool everythingIsTerrible = false;


        /// <summary>
        /// We apply filters as a coroutine when a inputfield is submitted,
        /// to allow the clear button a chance to clear it. We also use this
        /// right after a filter is applied, to resync morphs after a few frames.
        /// it is separate from other tasks so we can easily overwrite it.
        /// </summary>
        private IEnumerator submissionTask = null;

        private IEnumerator refreshTask = null;

        private bool isInitialized = false;
        private bool isUIInitialized = false;
        private bool isSubscribed = false;



        /// <summary>
        /// Re-subscribes to scene events if we are re-activating post-Initialization,
        /// and also scans for any new people.
        /// </summary>
        private void OnEnable()
        {
            // reset errors to try again.
            this.everythingIsTerrible = false;
            this.generalError = "";

            if (containingAtom.name != "CoreControl" || !string.IsNullOrEmpty(containingAtom.uid))
            {
                this.Log("MorphFilterPlus must be added as a session plugin.");
                this.generalError += " invalid plugin target, please set as a session plugin !";
                this.StopWorkingAndSulk();
            }            
            else if (this.isInitialized && !this.isSubscribed)
            {
                // if re-enabling after init: re-subscribe and re-scan.
                this.Subscribe();
                this.ScanForUpgradableAtoms();
            }
        }

        /// <summary>
        /// Unsubscribes from scene events and reverts any upgrades to people we had.
        /// called automatically by Monobehaviour API when disabled.
        /// </summary>
        private void OnDisable()
        {
            this.LogIf(this.isDebugMode, "MorphFilterPlus Disabling");
            this.Unsubscribe();
            foreach (var keyValuePair in this.upgradeLookup)
            {
                this.LogIf(this.isDebugMode, "MorphFilterPlus reverting " + (keyValuePair.Key != null ? keyValuePair.Key.name : "NULL"));
                keyValuePair.Value?.Revert();
            }
            this.upgradeLookup.Clear();
            this.tasks.Clear();
            this.submissionTask = null;
            this.refreshTask = null;
            MorphFilterPlus.IgnoredMorphs = null;
        }

        /// <summary>
        /// Initializes by subscribing to important scene events and 
        /// scanning for people already in the current scene. 
        /// Called automatically by VAM
        /// </summary>
        public override void Init()
        {
            this.LogIf(this.isDebugMode, "MorphFilterPlus.Init()");
            base.Init();
            if (containingAtom.name != "CoreControl" || !string.IsNullOrEmpty(containingAtom.uid))
            {
                this.Log("MorphFilterPlus must be added as a session plugin.");
                this.generalError += " invalid plugin target, please set as a session plugin !";
                this.StopWorkingAndSulk();
            }
            else
            {
                if (!this.isInitialized)
                {
                    // our own script stuff we only want to apply once
                    // regardless of how many times VAM calls Init()
                    this.Subscribe();
                    this.ScanForUpgradableAtoms();
                    this.isInitialized = true;
                }

                // the VAM-related stuff we'll apply every time Init() is called,
                // because it might call it multiple times on purpose.
                this.debugBoolJSON = new JSONStorableBool("Debug Logging", false);
                this.RegisterBool(this.debugBoolJSON);

                this.liveUpdateBoolJSON = new JSONStorableBool("Live Filtering", true);
                this.RegisterBool(this.liveUpdateBoolJSON);

                this.convertSpacesBoolJSON = new JSONStorableBool("Convert Spaces to '+'", false);
                this.RegisterBool(this.convertSpacesBoolJSON);

                this.minCharacterFloatJSON = new JSONStorableFloat("Live Filter Min Characters", 3.0f, 1f, 8f);
                this.RegisterFloat(this.minCharacterFloatJSON);
            }
        }

        /// <summary>
        /// Initializes the plugin's UI, automatically called by VAM.
        /// </summary>
        public override void InitUI()
        {
            base.InitUI();
            if (this.UITransform != null && this.isInitialized)
            {
                this.CreateToggle(this.debugBoolJSON, true);
                this.CreateToggle(this.liveUpdateBoolJSON);
                this.CreateToggle(this.convertSpacesBoolJSON);
                this.CreateSlider(this.minCharacterFloatJSON);
            }
        }

        /// <summary>
        /// Starts subscription to VAM's atom and scene events.
        /// these events will let us respond to loading a scene or atom.
        /// We call this on Init() and on OnEnable()
        /// </summary>
        private void Subscribe()
        {
            if (!this.isSubscribed)
            {
                SuperController.singleton.onAtomAddedHandlers += this.OnAtomAdded;
                SuperController.singleton.onAtomRemovedHandlers += this.OnAtomRemoved;
                SuperController.singleton.onSceneLoadedHandlers += this.OnSceneLoaded;
                SuperController.singleton.onSubSceneLoadedHandlers += this.OnSubSceneLoaded;

                this.isSubscribed = true;
            }
        }

        /// <summary>
        /// Ends subscription to VAM's atom and scene events.
        /// we unsubscribe on disable.
        /// </summary>
        private void Unsubscribe()
        {
            // since we unsub on disable, this gets called when we are quitting.
            // for all we know at that point the supercontroller is destroyed,
            // so we check first to avoid a null reference.
            // (We might be quitting but we can still be nice about it)
            if (SuperController.singleton != null)
            {
                SuperController.singleton.onAtomAddedHandlers -= this.OnAtomAdded;
                SuperController.singleton.onAtomRemovedHandlers -= this.OnAtomRemoved;
                SuperController.singleton.onSceneLoadedHandlers -= this.OnSceneLoaded;
                SuperController.singleton.onSubSceneLoadedHandlers -= this.OnSubSceneLoaded;
                this.isSubscribed = false;
            }

        }

        /// <summary>
        /// Iterates the Scene's Atoms to see if there are any new people.
        /// we call this when we don't know if there are, like when
        /// becoming enabled, or when loading a new scene or subscene.        
        /// </summary>
        public void ScanForUpgradableAtoms()
        {
            if (!this.everythingIsTerrible)
            {
                this.LogIf(this.isDebugMode, "MorphFilterPlus.ScanForPeople()");

                // GetSceneAtoms() is built into VAM's MVRScript API, 
                // and fetches the scene's root objects from the Supercontroller.
                // it's nice and fast, so we iterate that for people instead of 
                // using a FindObjectOfType<> or GetComponentsInChildren<> which are slower.

                foreach (var atom in this.GetSceneAtoms())
                {
                    switch(atom.type)
                    {
                        case AtomType.Person:
                            this.UpgradePerson(atom);
                            break;
                        // in previous versions we also modified subscenes
                        // but that is no longer necessary.
                    }
                    
                }
            }
        }


        /// <summary>
        /// Begins upgrading the person's input fields as a coroutine.
        /// We only make it a coroutine so we can yield a couple frames,
        /// to ensure the person is fully prepared for filter greatness.
        /// </summary>        
        public void UpgradePerson(Atom person)
        {
            // we only upgrade people that don't already have upgrades.
            // also we don't do it if stuff broke causing everything to be terrible.
            if (!this.upgradeLookup.ContainsKey(person) && !this.everythingIsTerrible)
            {
                this.LogIf(this.isDebugMode, $"MorphFilterPlus upgrading a new person named '{person.name}'");

                this.upgradeLookup[person] = new PersonUpgrade(person);
                this.tasks.Add(UpgradePersonProcess(person));
            }
            else if (this.isDebugMode && this.upgradeLookup.ContainsKey(person))
            {
                this.Log($"MorphFilterPlus already has an entry for person '{person.name}', ignoring.");
            }
        }


        /// <summary>
        /// This is the main process that results in 
        /// converting a standard filter input to our MorphFilterPlus input.
        /// we add it to our task list via UpgradePerson() when we find a new one.
        /// </summary>
        private IEnumerator UpgradePersonProcess(Atom person)
        {
           
            // now we start by waiting a few frames for the character to get settled in.
            // this is ugly but it's safe, so i'ma do it.
            // I don't even know if I really need to. I'm a radical.
            // radically cautious.
            yield return null;
            yield return null;
            yield return null;
            
            // ANYTHING could have happened in those 3 frames. Maybe our person's destroyed now.
            // we check that they exist and that we still care about them before continuing.
            if (person != null && this.upgradeLookup.ContainsKey(person) && !this.everythingIsTerrible)
            {                
                PersonUpgrade upgrade = this.upgradeLookup[person] as PersonUpgrade;

                // time to prepare for the magic. We fetch the person's male and female morph input fields,
                // using whatever named transform hierarchy path is accurate.
                Transform femaleMorphUIParent = person.transform.Find(k_femaleMorphUIPath);
                if (femaleMorphUIParent == null) femaleMorphUIParent = person.transform.Find(k_femaleMorphUIPathAlt);

                Transform maleMorphUIParent = person.transform.Find(k_maleMorphUIPath);
                if (maleMorphUIParent == null) maleMorphUIParent = person.transform.Find(k_maleMorphUIPathAlt);

                // schooled C# developers may note that i didn't use null propogation above.
                // schooled Unity developers will tell you that unity's Objects will LIE if you do. the Objects are LIARS

                #region create new filters to intercept input data
                // with InputFields in hand, we convert them to better ones.
                // our trick is duplicating the existing fields, disabling the old ones,
                // and having the new ones point events at us, so we intercept the input.
                // keeping the old ones around makes it easy to revert when 
                // either this plugin or the person is removed.
                #endregion

                if (femaleMorphUIParent != null && maleMorphUIParent!= null)
                {
                    try
                    {

                        upgrade.OldFieldMale = maleMorphUIParent.Find(k_filterInputField).GetComponent<InputField>();
                        upgrade.OldFieldFemale = femaleMorphUIParent.Find(k_filterInputField).GetComponent<InputField>();

                        upgrade.NewFieldMale = this.CreateFilterField(person, upgrade.OldFieldMale);
                        upgrade.NewFieldFemale = this.CreateFilterField(person, upgrade.OldFieldFemale);

                        upgrade.UIMale = maleMorphUIParent.Find(k_maleMorphUIControl).GetComponent<GenerateDAZMorphsControlUI>();
                        upgrade.UIFemale = femaleMorphUIParent.Find(k_femaleMorphUIControl).GetComponent<GenerateDAZMorphsControlUI>();

                        #region Add UI Modifications and Listeners
                        // VAM will rebuild the UI when the morph panel toggles or pop up selections are changed.
                        // currently, we filter things by telling them to be invisible only while the UI is about to be built,
                        // and then reverting them to visible again for general compatibility. This means we need to know
                        // when the UI is about to be rebuilt again, such as when these UI elements are being interacted with,
                        // that way we can adjust the visibility again. For some things, it's most efficient to just subscribe
                        // to the interaction event and re-apply the filter. For other things it's better to create our own
                        // UI element to intercept the interaction.
                        #endregion

                        // for the 'only show' toggles like latest, active, and favorites, we create a "UIBridge" that intercepts the click
                        // before forwarding it on to the toggle. This guarantees we react to the click before VAM does, as subscription order is not guaranteed.
                        upgrade.AddBridge(new ToggleBridge(femaleMorphUIParent.Find(k_onlyActiveToggle).GetComponent<Toggle>(), initialValue: false, hideOldElement: true, forwardEvents: true));
                        upgrade.AddBridge(new ToggleBridge(femaleMorphUIParent.Find(k_onlyFavToggle).GetComponent<Toggle>(), initialValue: false, hideOldElement: true, forwardEvents: true));
                        upgrade.AddBridge(new ToggleBridge(femaleMorphUIParent.Find(k_onlyLatestToggle).GetComponent<Toggle>(), initialValue: true, hideOldElement: true, forwardEvents: true));
                        upgrade.AddBridge(new ToggleBridge(maleMorphUIParent.Find(k_onlyActiveToggle).GetComponent<Toggle>(), initialValue: false, hideOldElement: true, forwardEvents: true));
                        upgrade.AddBridge(new ToggleBridge(maleMorphUIParent.Find(k_onlyFavToggle).GetComponent<Toggle>(), initialValue: false, hideOldElement: true, forwardEvents: true));
                        upgrade.AddBridge(new ToggleBridge(maleMorphUIParent.Find(k_onlyLatestToggle).GetComponent<Toggle>(), initialValue:true, hideOldElement: true, forwardEvents: true));
                        upgrade.BridgeEvent -= this.OnMorphPanelBridgeEvent;
                        upgrade.BridgeEvent += this.OnMorphPanelBridgeEvent;

                        // when selecting a category or type, we do the simpler approach of subscribing to the selection event.
                        // however this requires us to fully re-apply the filter, resync, and restore UI after a few frames,
                        // since VAM is going to regenerate the UI before we have a chance to filter. it adds lag but I'm not sure we want to
                        // add a bridge for every button in every dropdown.
                        upgrade.SetSubscription(maleMorphUIParent.Find(k_showTypePopup).GetComponent<UIPopup>().onValueChangeUnityEvent, true);
                        upgrade.SetSubscription(maleMorphUIParent.Find(k_maleCategoryPopup).GetComponent<UIPopup>().onValueChangeUnityEvent, true);
                        upgrade.SetSubscription(femaleMorphUIParent.Find(k_showTypePopup).GetComponent<UIPopup>().onValueChangeUnityEvent, true);
                        upgrade.SetSubscription(femaleMorphUIParent.Find(k_femaleCategoryPopup).GetComponent<UIPopup>().onValueChangeUnityEvent, true);                        
                        upgrade.BaseUIEvent -= this.OnMorphPanelUIEvent;
                        upgrade.BaseUIEvent += this.OnMorphPanelUIEvent;

                    }
                    catch (Exception e)
                    {
                        this.Log("MorphFilterPlus Failed to 'upgrade' person: " + e);
                        generalError += " upgrade exception !";
                    }
                }
                else
                {
                    this.generalError += " missing input fields !";
                    this.Log("Something went wrong and MorphFilterPlus couldn't find the normal input field. Did you do something?" +
                        " You changed something, didn't you? What have you done? WHAT DID YOU DO", error: true);
                }


                if (!this.hasError)
                {
                    #region get morph bank components from geometry children
                    // next, in order to actually do anything with the filter, we need to be able to
                    // access the morphs. Morphs are stored in components on child game objects of the person,
                    // one game object for each category of morph. These in turn are managed by 5 MorphBanks; 
                    // one for each gender's base and genital morphs, and an extra genital morph group for males.
                    // These 5 overarching MorphBanks are also components on game objects-- they're the parents of the categories.
                    // they keep track of all of the morphs of their child subbanks.
                    // we are going to take advantage of that when we filter by using those parent morphbanks.

                    // rather than doing a GetComponent<> every time we filter, we're going to
                    // find and cache the morph banks now as part of the person's upgrade profile.
                    #endregion

                    upgrade.MorphBanks = new DAZMorphBank[MorphFilterPlus.morphBankPaths.Length];
                    for (int i = 0; i < MorphFilterPlus.morphBankPaths.Length; i++)
                    {
                        upgrade.MorphBanks[i] = GetMorphBank(person, MorphFilterPlus.morphBankPaths[i]);
                        if (this.hasError) break;
                    }
                    
                    if (!this.hasError)
                    {
                        // we do some redundant initialization / resyncing 
                        // to remove dynamic generated morphs from visibility;
                        // though these lines might not be the reason for success and may be removed later.
                        upgrade.ResyncMorphs();
                        this.tasks.Add(this.DelayedAction(5, () => { upgrade.RefreshUI(); }));
                    }
                }

            }// end do upgrade
            

            if (this.hasError)
            {
                this.Log("MorphFilterPlus encountered errors: " + this.generalError, error: true);

                // if at any point we ran into a major error like a missing transform,
                // then this plugin is just not going to work, and we should give up and disable.
                this.StopWorkingAndSulk();
            }

        }



        /// <summary>
        /// Finds a child transform of the person according to a transform path,
        /// and fetches the morph bank component from it, if possible.
        /// Flags an error if it can't find the transform or morph bank.
        /// </summary>                
        private DAZMorphBank GetMorphBank(Atom person, string path)
        {
            DAZMorphBank ret_bank = null;
            Transform bankTransform = person.transform.Find(path);
            if (bankTransform != null) ret_bank = bankTransform.GetComponent<DAZMorphBank>();

            if (ret_bank == null) this.generalError += $" missing morphbank at path '{path}' !";

            return ret_bank;
        }

        /// <summary>
        /// Performs the legwork of duplicating the vanilla VAM morph search InputField 
        /// and intercepting event from the new one (and turning it blue). 
        /// flags an error if it can't find important elements.
        /// </summary>        
        private InputField CreateFilterField(Atom person, InputField oldField)
        {
            InputField ret_newField = null;
            try
            {
                ret_newField = Instantiate(oldField, oldField.transform.parent);

                // i'm not 100% sure if there's some layout stuff happening, so to preserve any
                // we make sure to plop the new field at the same spot in the hierarchy.
                ret_newField.transform.SetSiblingIndex(oldField.transform.GetSiblingIndex());

                // we don't want our duplicate field to send events to VAM's stuff,
                // so we scrub the events it might have...
                ret_newField.onValueChanged.RemoveAllListeners();
                ret_newField.onEndEdit.RemoveAllListeners();

                // ...and we add our own methods as listeners instead.
                ret_newField.onValueChanged.AddListener((value) => this.OnFilterChange(person, ret_newField));
                ret_newField.onEndEdit.AddListener((value) => this.OnFilterSubmit(person, ret_newField));

                // the clear button is duplicated too as a child
                var clearButton = ret_newField.GetComponentInChildren<Button>();
                clearButton.onClick.RemoveAllListeners();
                clearButton.onClick.AddListener(() => this.OnFilterClear(person, ret_newField));

                // add our signature MorphFilterBlue(TM)
                ret_newField.GetComponent<Image>().color = MorphFilterPlus.mintyFreshColor;

                // we also need to know when the filter gets enabled so we can re-apply it,
                // since the UI refreshes every time it's opened. this is a little inelegant imo
                // but it will do the trick for now. since it's part of the field
                // and the field is destroyed on reversion, we don't need to worry about lifecycle.
                var listener = ret_newField.gameObject.AddComponent<StateListener>();
                listener.Owner = person;
                listener.EnabledEvent += this.OnMorphUIEnabled;

                // and we're done. wrap up by simply turning off the original input field
                oldField.gameObject.SetActive(false);
            }
            catch (Exception e)
            {
                this.generalError += "MorphFilterPlus.CreateFilterField() exception: " + e;
            }
            return ret_newField;
        }


        /// <summary>
        /// when a category or morph type selection is made, we are subscribing directly to it.
        /// this means we aren't reliably able to perform an action before the UI updates.
        /// We therefore commit a re-application of all parts of the filter, after a frame delay.
        /// </summary>        
        private void OnMorphPanelUIEvent(object sender, EventArgs args)
        {            
            // this is costly since it's basically re-applying the UI right after the UI applied itself.
            // the alternative is creating a bridge UI for every button in a pop-up, which we might do later,
            // but for now the simplicity of subscribing to the change outweighs that performance boost.
            var upgrade = (PersonUpgrade)sender;
            InputField visibleField = upgrade.GetVisibleField();

            this.refreshTask = null;
            // we delay a frame so the timing of our submission vs VAM's UI rebuild is a guaranteed thing.
            this.submissionTask = this.DelayedAction(1, ()=>this.ApplyFilter(upgrade.Owner, visibleField));

        }

        /// <summary>
        /// when a "show only" toggle bridge is clicked, we immediately re-filter and re-sync the morphs. 
        /// this is guaranteed to occur before VAM hears the event, at which point VAM will
        /// rebuild the UI. We add a delayed action to restore default visibility after the ui is rebuilt.
        /// </summary>        
        private void OnMorphPanelBridgeEvent(object sender, EventArgs args)
        {
            var upgrade = (PersonUpgrade)sender;
            InputField visibleField = upgrade.GetVisibleField();
            if (visibleField != null)
            {
                this.InitIgnoreList(upgrade);
                this.submissionTask = null;
                upgrade.FilterMorphs(visibleField, this.convertSpacesBoolJSON.val);
                upgrade.ResyncMorphs(visibleField);
                this.refreshTask = this.DelayedAction(5, () => upgrade.RestoreMorphs(visibleField));
            }
        }

        /// <summary>
        /// when a morph ui panel is enabled we ensure our filter is applied before
        /// the UI finishes refreshing. We subscribed this to a StateListener behaviour
        /// added to our new input fields when we created them.
        /// </summary>        
        private void OnMorphUIEnabled(object sender, EventArgs args)
        {
            try
            {
                var listener = (StateListener)sender;
                var field = listener.GetComponent<InputField>();
                AtomUpgrade upgrade;
                if (field != null && this.upgradeLookup.TryGetValue(listener.Owner, out upgrade))
                {
                    // when enabled, the UI panel does a delayed category refresh, destroying its children
                    // then rebuilding the morphs after a few frames. We re-filter our visibility
                    // and resync the morphs immediately so the UI builds our filtered display,
                    // then delay a restoration of default values after the ui is rebuilt.
                    this.submissionTask = null;
                    var personUpgrade = (PersonUpgrade)upgrade;
                    personUpgrade.FilterMorphs(field, this.convertSpacesBoolJSON.val);
                    personUpgrade.ResyncMorphs(field);
                    this.refreshTask = this.DelayedAction(5, () => personUpgrade.RestoreMorphs(field));
                }
                else
                {
                    this.Log("MorphFilterPlus was unable to handle an enable event due to missing components.", error: true);
                }
            }
            catch (Exception e)
            {
                this.Log("MorphFilterPlus was unable to respond to a ui enable event: " + e, error: true);
            }

        }

        /// <summary>
        /// A general-use action coroutine we can add as a task
        /// to delay an arbitrary number of frames before doing something.        
        /// </summary>        
        /// <remarks>
        /// we use this to scan for people after loading a scene, for example,
        /// so that atoms are given a chance to initialize and any other plugins
        /// that might be messing with initialization are given a chance to do it.
        /// </remarks>
        private IEnumerator DelayedAction(int frameDelay, Action action)
        {
            int framesRemaining = frameDelay;
            while (!this.hasError && --framesRemaining > 0)
            {
                yield return null;
            }

            if (!this.hasError)
            {
                action();
            }
        }

        /// <summary>
        ///  We parse the text from the filter and apply it, IF live mode is active.         
        ///  We subscribed this to the input change event of the InputField we created
        ///  in our UpgradeField(..) method. 
        /// </summary>        
        private void OnFilterChange(Atom person, InputField field)
        {
            // we don't do anything for non-live submission mode or regex
            // and we wait for the characters to pass our threshold.
            if ((this.liveUpdateBoolJSON.val && !field.text.StartsWith("@")) && (field.text.Length >= Mathf.Floor(this.minCharacterFloatJSON.val) || field.text.Length == 0))
            {                
                this.submissionTask = null;
                this.refreshTask = null;
                this.ApplyFilter(person, field);
            }

        }


        /// <summary>
        /// Clears the text of a given field, causing OnFilterChange() to be called.
        /// if live update is false, this immediately restores morphs.        
        /// We subscribed this to the new clear button's OnClick event in our UpgradeField(..) method
        /// </summary>
        private void OnFilterClear(Atom person, InputField field)
        {
            // setting the text will prompt OnFilterChange if it wasn't already empty,
            // so we don't have to do anything else here for live-mode.
            field.text = "";
            if (this.submissionTask == null && !this.hasError)
            {
                // if we have NOT started to submit the input, then let's
                // restore everything and refresh the UI.

                this.refreshTask = null;                        
                AtomUpgrade upgrade;
                if (this.upgradeLookup.TryGetValue(person, out upgrade))
                {
                    this.InitIgnoreList(upgrade);
                    var personUpgrade = upgrade as PersonUpgrade;
                    personUpgrade.RestoreMorphs(field, resync:true);
                    personUpgrade.RefreshUI(field);
                }
            }

            // if we HAVE started to submit something, then it will
            // detect the empty text when it finishes, and clear the filter anyway,
            // so we don't need to account for that condition.

            if (this.hasError)
            {
                this.StopWorkingAndSulk();
            }


        }


        /// <summary>
        /// Applies the inputfield as a filter IF live update is false, or we are in regex mode.        
        /// </summary>
        /// <remarks>
        /// This assigns our submission to a task we update each frame, as careful timing is necessary.
        /// Detailed comments are available within for more info.
        /// </remarks>
        private void OnFilterSubmit(Atom person, InputField field)
        {            
            // we only use submission-based filtering if our live update toggle is off.
            if ((!this.liveUpdateBoolJSON.val || field.text.StartsWith("@") || field.text.Length < this.minCharacterFloatJSON.val) && !this.hasError)
            {                
                #region create a carefully timed task to allow clear button events.

                // here we must be clever.
                // the submit method is called when we hit enter OR click away from the field.
                // this means it's called when we click the Clear button (because we're clicking away).
                // if we don't account for that we'll end up applying the text filter 
                // and then clearing it immediately, which will both lag and possibly break some 
                // hidden complexity within VAM's morph or UI management.

                // so what we do instead is start a submission task where we wait until a 
                // few frames after the mouse has been released before applying the text of the inputfield.
                // this gives us a chance to clear it without applying it first.                

                // this task is a cached member we replace when necessary.
                // we don't start it as a coroutine but instead step through it in update.
                #endregion

                this.submissionTask = this.SubmissionProcess(person, field);
            }            
        }

        /// <summary>
        /// The coroutine we step through to submit a filter that wasn't live updated,
        /// while ensuring we can click the clear button without double-submitting.
        /// </summary>                
        private IEnumerator SubmissionProcess(Atom person, InputField field)
        {
            // in case the mouse doesn't register as down this specific frame, give it a sec.            
            yield return null;

            // if your mouse/keyboard is stuck we don't want to loop forever, even if it is a passive loop.
            int frameLimit = 600;
            while (Input.anyKey && !this.everythingIsTerrible && --frameLimit > 0)
            {
                // we use anykey to cover all input bases even though it's probably a mouse button.
                // we are waiting for the mouse to be released so we know a clear button click
                // will or will not be happening
                yield return null;
            }

            if (frameLimit <= 0)
            {
                this.Log("MorphFilterPlus couldn't finish filter submission; A key or mouse was detected for 600 frames.", error: true);
                this.generalError += " submission stalled !";
            }
            else
            {
                // once the mouse is released, give it a frame so that if it was clicking
                // the clear button, it will definitely have called the clear event before we apply.
                yield return null;

                // now that no input is down and the UI did its thing, we apply.
                this.ApplyFilter(person, field);
            }
        }

        private void InitIgnoreList(AtomUpgrade upgrade)
        {

            if (MorphFilterPlus.IgnoredMorphs == null && !this.hasError && upgrade is PersonUpgrade)
            {
                var personUpgrade = (PersonUpgrade)upgrade;
                // if this is the first time we've tried applying a filter, we'll populate our ignored morphs cache.
                // we waited this long because some morphs seem to be generated or hidden dynamically on selection,
                // and we would have missed them in the upgrade step. here we know for sure the character has been selected.  
                MorphFilterPlus.IgnoredMorphs = new HashSet<string>();
                foreach (DAZMorphBank bank in personUpgrade.MorphBanks)
                {
                    foreach (DAZMorph morph in bank.morphs)
                    {
                        if (!morph.visible) MorphFilterPlus.IgnoredMorphs.Add(morph.uid);
                    }
                }
                
                this.LogIf(this.isDebugMode, $"MorphFilterPlus is ignoring {MorphFilterPlus.IgnoredMorphs.Count} morphs.");

                //// efforts to prevent eyelid morphs from appearing in the first search have led to
                //// an initial restore / refresh being called. I think it might have resolved it.
                //upgrade.RestoreMorphs();
                //upgrade.RefreshUI();
            }
        }

        /// <summary>
        /// Filters the morphs visibility then resyncs and rebuilds the UI, before
        /// restoring original visibility settings to preserve compatibility.        
        /// It's called when an input field changes if live update is true, 
        /// or after a submission event if live update is false.
        /// </summary>        
        private void ApplyFilter(Atom person, InputField field)
        {
            
            AtomUpgrade baseUpgrade;
            if (!this.hasError && this.upgradeLookup.TryGetValue(person, out baseUpgrade))
            {
                PersonUpgrade upgrade = baseUpgrade as PersonUpgrade;
                this.InitIgnoreList(upgrade);

                try
                {
                    upgrade.FilterMorphs(field, this.convertSpacesBoolJSON.val);
                    upgrade.ResyncMorphs(field);
                    upgrade.RefreshUI(field);

                    #region wait a few frames to restore visibility settings
                    // after applying the filter we want to restore factory default visibility,
                    // basically undoing the visibility setting we just did, so that it doesn't
                    // interfere with saves or other systems. However, VAM takes a few frames to rebuild UI.

                    // so we account for that by waiting a few EXTRA frames before restoring it,
                    // otherwise the ui would be rebuilt with the restored settings and
                    // we wouldn't have filtered at all.
                    #endregion
                    this.refreshTask = this.DelayedAction(5, () => upgrade.RestoreMorphs(field, resync: true));
                }
                catch (Exception e)
                {
                    this.Log("MorphFilterPlus exception applying filter: " + e);
                    this.generalError += " failed to apply filter !";
                }
            }


            if (this.hasError)
            {
                this.StopWorkingAndSulk();
            }

        }

        /// <summary>
        /// Called by Unity every frame automatically as part of the MonoBehavior API
        /// we inherit. We use it to step through our coroutines each frame. 
        /// Update has overhead so it's a tiny bit wasteful for this. I don't want to use it.
        /// but I really REALLY don't like StartCoroutine, so I'm going to anyway for now.
        /// </summary>
        private void Update()
        {
            if (!this.everythingIsTerrible)
            {
                for (int i = 0; i < this.tasks.Count; i++)
                {
                    if (!this.tasks[i].MoveNext())
                    {
                        this.tasks.RemoveAt(i);
                        i--;
                    }
                }

                if (this.submissionTask != null)
                {
                    if (!this.submissionTask.MoveNext()) this.submissionTask = null;
                }

                if (this.refreshTask != null)
                {
                    if (!this.refreshTask.MoveNext()) this.refreshTask = null;
                }
            }
        }

        /// <summary>
        /// Responds to newly created atoms by checking if they're a person
        /// and upgrading them if so. This is one of our event subscriptions in OnEnable()
        /// </summary>        
        private void OnAtomAdded(Atom atom)
        {
            switch (atom.type)
            {
                case AtomType.Person:
                    this.UpgradePerson(atom);
                    break;
            }
        }

        /// <summary>
        /// Responds to the removal of an atom by checking if they're a person
        /// and reverting their upgrade if so. This is one of our event subscriptions in OnEnable()
        /// </summary>
        /// <param name="atom"></param>
        private void OnAtomRemoved(Atom atom)
        {
            AtomUpgrade upgrade;
            if (atom != null && this.upgradeLookup.TryGetValue(atom, out upgrade))
            {
                this.LogIf(this.isDebugMode, "MorphFilterPlus Reverting mods for removed atom: " + atom.name);
                upgrade?.Revert();
                this.upgradeLookup.Remove(atom);
            }

        }

        /// <summary>
        /// Responds to the VAM scene being loaded by scanning for people present,
        /// after a brief delay to allow any weird atom things to be settled.
        /// </summary>
        private void OnSceneLoaded()
        {
            IEnumerator delayedScan = this.DelayedAction(frameDelay: 5, action: this.ScanForUpgradableAtoms);
            this.tasks.Add(delayedScan);
        }


        /// <summary>
        /// Responds to a VAM subscene being loaded by scanning for new people,
        /// after a brief delay to allow any weird atom things to be settled.
        /// </summary>        
        private void OnSubSceneLoaded(SubScene scene)
        {
            IEnumerator delayedScan = this.DelayedAction(frameDelay: 5, action: this.ScanForUpgradableAtoms);
            this.tasks.Add(delayedScan);
        }


        /// <summary>
        /// Sets the flag implicating that something is critically wrong,
        /// and disables the script. This is called when we encounter an issue
        /// that would prevent the plugin from working, such as missing transforms.
        /// Since it disables itself, it results in the reversion of any existing upgrades.
        /// </summary>
        private void StopWorkingAndSulk()
        {
            if (!this.everythingIsTerrible)
            {

                this.Log($"MorphFilterPlus is shutting itself down with these errors: {this.generalError} \nThings didn't go as planned, and MorphFilterPlus has decided to sulk. " +
                    "You can try re-enabling or reloading it if you think it's a fluke.", error: true);

                this.everythingIsTerrible = true;
                this.generalError = "";

                // we don't want to just disable the script, we want to show it as disabled
                // in VAM's plugin UI by turning off the toggle. Doing so will result
                // in a callback that disables it, calls OnDisable, and reverts our upgrades as desired.
                // however if there is a nicer looking way to do this, let me (GossamerVR) know.                
                if (this.enabledJSON != null && this.enabledJSON.toggle != null)
                {
                    this.enabledJSON.toggle.isOn = false;
                }


            }
        }



        private void LogIf(bool condition, string str, bool error = false)
        {
            if (condition) this.Log(str);
        }

        private void Log(string str, bool error = false)
        {
            if (error) SuperController.LogError(str);
            else SuperController.LogMessage(str);
        }



    }
}