namespace GossamerVR.MorphFilterPlus
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using UnityEngine;
    using UnityEngine.UI;
    using UnityEngine.Events;
    using UnityEngine.EventSystems;

    /// <summary>
    /// The base class of any modifications we make to atoms;
    /// these objects will include references to any new or changed objects,
    /// and handle their own removal. Currently we only modify person atoms.
    /// (previously person and subscenes)
    /// </summary>
    public abstract class AtomUpgrade
    {
        /// <summary>
        /// published when a UI or UI bridge element is clicked or toggled.
        /// </summary>
        public event EventHandler BaseUIEvent;
        public event EventHandler BridgeEvent;

        public Atom Owner { get; protected set; }

        // we handle listening to UI elements through either direct subscription to ui events,
        // or through 'bridges' that can guarantee our actions occur first.
        private List<UnityEventBase> uiEvents = new List<UnityEventBase>();
        private List<IUIElementBridge> uiBridges = new List<IUIElementBridge>();        
        public int BridgeCount => this.uiBridges.Count;

        
        /// <summary>
        /// Adds a bridge to this upgrade to keep track of,
        /// so that it is removed when this upgrade is reverted.
        /// </summary>            
        public void AddBridge(IUIElementBridge bridge)
        {
            if (bridge != null && !this.uiBridges.Contains(bridge))
            {
                bridge.InteractEvent += this.OnBridgeInteract;
                bridge.RemovalEvent += this.OnBridgeRemoved;
                this.uiBridges.Add(bridge);
            }
        }

        /// <summary>
        /// adds or removes a subscription to a unity event, typically a UI element event.
        /// maintains a reference to the subscription so we can unsubscribe on removal.
        /// </summary>        
        public void SetSubscription(UnityEventBase evt, bool subscribe)
        {
            if (!subscribe || !this.uiEvents.Contains(evt))
            {
                if (evt is UnityEvent)
                {
                    var simpleEvent = (UnityEvent)evt;
                    if (subscribe) simpleEvent.AddListener(this.OnUIEvent);
                    else simpleEvent.RemoveListener(this.OnUIEvent);
                }
                else if (evt is UnityEvent<string>)
                {
                    var stringEvent = (UnityEvent<string>)evt;
                    if (subscribe) stringEvent.AddListener(this.OnTypedEvent<string>);
                    else stringEvent.RemoveListener(this.OnTypedEvent<string>);
                }
                else if (evt is UnityEvent<bool>)
                {
                    var boolEvent = (UnityEvent<bool>)evt;
                    if (subscribe) boolEvent.AddListener(this.OnTypedEvent<bool>);
                    else boolEvent.RemoveListener(this.OnTypedEvent<bool>);
                }
                else
                {
                    SuperController.LogMessage("MorphFilterPlus was told to listen to an event type it doesn't support: " + evt.ToString());
                }
                if (subscribe) this.uiEvents.Add(evt);
                else this.uiEvents.Remove(evt);
            }
        }

        // we aren't doing anything with typed events at the moment,
        // just publishing the same interaction event as everything else.
        protected virtual void OnTypedEvent<T>(T value) => this.OnUIEvent();
        protected virtual void OnUIEvent()
        {
            this.BaseUIEvent?.Invoke(this, EventArgs.Empty);
        }
        protected virtual void OnBridgeInteract(object sender, EventArgs args)
        {
            this.BridgeEvent?.Invoke(this, EventArgs.Empty);
        }

        /// <summary>
        /// unsubs the bridge and removes it from tracking,
        /// this is called when Remove() is called on the bridge itself.
        /// (it is subscribed to the bridge's RemovalEvent)
        /// </summary>        
        protected virtual void OnBridgeRemoved(object sender, EventArgs args)
        {
            var bridge = (IUIElementBridge)sender;
            bridge.InteractEvent -= this.OnBridgeInteract;
            bridge.RemovalEvent -= this.OnBridgeRemoved;
            this.uiBridges.Remove(bridge);
        }

        /// <summary>
        /// Unsubscribes and removes all elements of the upgrade and destroys
        /// all modification parts or bridges. Called when removing the plugin
        /// or atom, so that the atom's natural state is restored.
        /// </summary>
        public virtual void Revert()
        {   

            List<IUIElementBridge> elementListCopy = new List<IUIElementBridge>(this.uiBridges);
            elementListCopy.ForEach(x => x.Remove());
            this.uiBridges.Clear();

            List<UnityEventBase> eventListCopy = new List<UnityEventBase>(this.uiEvents);
            eventListCopy.ForEach(x => this.SetSubscription(x, false));
            this.uiEvents.Clear();

            this.Owner = null;
        }
    }
}