namespace GossamerVR.MorphFilterPlus
{
    using System;    
    using System.Collections;
    using System.Collections.Generic;
    using System.Text.RegularExpressions;
    using UnityEngine;
    using UnityEngine.UI;
    using UnityEngine.Events;
    using UnityEngine.EventSystems;
    
    /// <summary>
    /// Handles references and effects for a specific person and their UI elements,
    /// Persons have morph input fields, ui, and banks we track in addidion to the 
    /// base save appearance button.
    /// </summary>
    public class PersonUpgrade : AtomUpgrade
    {

        public InputField OldFieldFemale;
        public InputField OldFieldMale;
        
        public InputField NewFieldFemale;
        public InputField NewFieldMale;

        

        // to speed up filtering and re-applying filters,
        // we cache our last search and last built filters,
        // so if the search hasn't changed we don't have to rebuild.
        private string lastAppliedStringMale;
        private List<List<Filter>> lastAppliedFiltersMale = new List<List<Filter>>();
        private string lastAppliedStringFemale;
        private List<List<Filter>> lastAppliedFiltersFemale = new List<List<Filter>>();

        public GenerateDAZMorphsControlUI UIFemale;
        public GenerateDAZMorphsControlUI UIMale;

        public DAZMorphBank[] MorphBanks;

        public PersonUpgrade(Atom owner) { this.Owner = owner; }


        /// <summary>
        /// An ease-of-use method allowing a fetch of the currently visible upgrade input field,
        /// primarily used for creating actions to resync or refresh the relevant UI.
        /// </summary>        
        public InputField GetVisibleField()
        {
            if (this.NewFieldMale != null && this.NewFieldMale.gameObject.activeInHierarchy)
            {
                return this.NewFieldMale;
            }
            else if (this.NewFieldFemale != null && this.NewFieldFemale.gameObject.activeInHierarchy)
            {
                return this.NewFieldFemale;
            }
            return null;
        }

        ///// <summary>
        ///// Dynamically determines which input field to apply based on 
        ///// which is visible in the scene currently, and then applies it.
        ///// </summary>
        //public void FilterMorphs(bool visibleOnly = false)
        //{
        //    if (this.NewFieldFemale != null && (!visibleOnly || this.NewFieldFemale.gameObject.activeInHierarchy))
        //    {
        //        this.FilterMorphs(this.NewFieldFemale);
        //    }
        //    else if (this.NewFieldMale != null && (!visibleOnly || this.NewFieldMale.gameObject.activeInHierarchy))
        //    {
        //        this.FilterMorphs(this.NewFieldMale);
        //    }            
        //}

        /// <summary>
        /// Iterates all morphs relevant to a specific field, builds filters
        /// based on the field's input, and applies visibility settings
        /// to each morph based on whether they match the filters.
        /// This is intended to be a transient effect that is restored
        /// once the UI is up-to-date so that no invisibility affects
        /// saves or other systems. This method does not perform restoration.
        /// </summary>        
        public void FilterMorphs(InputField field, bool convertSpaces)
        {
            List<DAZMorphBank> banks = new List<DAZMorphBank>();

            // first we get the morph banks relevant to the field.
            if (field == this.NewFieldFemale)
            {
                banks.Add(this.MorphBanks[MorphFilterPlus.BankIndex.FemaleChar]);
                banks.Add(this.MorphBanks[MorphFilterPlus.BankIndex.FemaleGen]);
                this.FilterMethod(banks, this.NewFieldFemale, this.lastAppliedStringFemale, ref this.lastAppliedFiltersFemale, convertSpaces);
                this.lastAppliedStringFemale = this.NewFieldFemale.text;
            }
            else if (field == this.NewFieldMale)
            {
                banks.Add(this.MorphBanks[MorphFilterPlus.BankIndex.MaleChar]);
                banks.Add(this.MorphBanks[MorphFilterPlus.BankIndex.MaleGen]);
                banks.Add(this.MorphBanks[MorphFilterPlus.BankIndex.MaleGen2]);
                this.FilterMethod(banks, this.NewFieldMale, this.lastAppliedStringMale, ref this.lastAppliedFiltersMale, convertSpaces);
                this.lastAppliedStringMale = this.NewFieldMale.text;
            }            
        }

        /// <summary>
        /// The internal method by which morphs for a given set of banks are filtered.
        /// this method does the work of determinine if a cache should be used or building new filters,
        /// and then iterating the banks to apply the filters.
        /// </summary>        
        private void FilterMethod(List<DAZMorphBank> banks, InputField field, string lastString, ref List<List<Filter>> filters, bool convertSpaces)
        {

            if (string.IsNullOrEmpty(field.text.Trim()))
            {
                // if the filter is empty, we just restore all the morphs.
                this.RestoreMorphs(field, resync:false);
            }
            else if (field.text.StartsWith("@"))
            {
                // new in 3.0 starting with @ uses regular expressions.
                Regex regex = new Regex(field.text.Substring(1), RegexOptions.IgnoreCase);
                int matchCount = 0;
                foreach (DAZMorphBank bank in banks)
                {
                    foreach (DAZMorph morph in bank.morphs)
                    {

                        if (MorphFilterPlus.IgnoredMorphs.Contains(morph.uid)) continue;
                        try
                        {
                            morph.visible = regex.IsMatch(morph.lowerCaseResolvedDisplayName);
                        }
                        catch (Exception e)
                        {
                            SuperController.LogError($"MorphFilterPlus had an issue regex matching morph '{(morph == null ? "NULL" : morph.displayName)} : {e}");
                            morph.visible = true; // might as well show morphs we aren't sure about.                            
                        }
                    }
                }
            }
            else 
            {
                #region the filter starts by splitting into separate requirement sets

                // it gets tricky here but bear with me.
                // we have two tiers of "AND" (as of 3.0), the "AND" (&) and the "PLUS" (+). the plus
                // is treated like any other constraint but the AND and OR will create logical groups.                    
                // AND was added to give more control of the order of operations. It is applied first.

                // First we split the input text into multiple inputs split by the AND code.
                // these will end up being filter sets with their own rules, and all of them will have
                // to give the greenlight to any morph they're compared against.
                #endregion
                string filterText = convertSpaces ? field.text.Trim().Replace(" ","+") : field.text;
                if (field.text != lastString)
                {
                    // we only rebuild filters if the filter string has changed.
                    // since our filters param is a ref, this will overwrite the relevant list of filters for this field.
                    filters = new List<List<Filter>>();
                    string[] andSegments = filterText.Split(new[] { '&' }, StringSplitOptions.RemoveEmptyEntries);
                    foreach (string segment in andSegments)
                    {
                        List<Filter> optionalSet = new List<Filter>();

                        // to build a set of filters within each AND group, we first split each string into 
                        // OR segments. This will give us a final set of strings that will turn into filters.
                        // Since these are separated by OR, the morph only has to match one of them 
                        // for us to give the greenlight.
                        string[] orSegments = segment.Split(new[] { '%','/','|' }, StringSplitOptions.RemoveEmptyEntries);
                        foreach (var constraintCombo in orSegments)
                        {
                            optionalSet.Add(new Filter(constraintCombo.ToLower()));
                        }

                        filters.Add(optionalSet);
                    }
                }
                                
                // now that the filters are built, we are going to compare them 
                // to each of the morph display names, and update the morph visibility.                    
                /// using continue to avoid nesting in a few places here.
                bool match = true;
                foreach (DAZMorphBank bank in banks)
                {
                    foreach (DAZMorph morph in bank.morphs)
                    {

                        if (MorphFilterPlus.IgnoredMorphs.Contains(morph.uid))
                        {
                            continue;
                        }

                        // by default everything is visible if we have no filters.
                        match = filters.Count == 0;
                        if (match)
                        {
                            morph.visible = true;
                            continue;
                        }

                        // for the required set, we presume innocence until proven guilty.
                        match = true;
                        foreach (List<Filter> optionalSet in filters)
                        {
                            if (optionalSet.Count == 0) continue;

                            // for the optional set we presume guilt until proven innocent
                            match = false;
                            foreach (Filter filter in optionalSet)
                            {
                                match = filter.CompareString(morph.lowerCaseResolvedDisplayName);
                                if (match) break;
                            }

                            // once we've gone through the gamut of optional filters,
                            // if we haven't matched, the required set clearly isn't all met.
                            // so we don't need to look any more.
                            if (!match) break;
                        }

                        // whatever happened, we accept the result.
                        morph.visible = match;

                    }// end morph loop
                }// end banksy loop        
            }// end use filters condition

        }


        /// <summary>
        /// Unhides all morphs for this person except ignored ones.
        /// called when removing the plugin or person, as well as 
        /// before saving to ensure the morphs can be saved.
        /// </summary>
        public void RestoreMorphs(bool resync = true)
        {
            if (this.MorphBanks != null && MorphFilterPlus.IgnoredMorphs != null)
            {
                foreach (var bank in this.MorphBanks)
                {
                    this.RestoreBank(bank);
                }
            }
            if (resync) this.ResyncMorphs();
        }


        /// <summary>
        /// Restores morphs only relevant to the input field (fem or male)
        /// </summary>            
        public void RestoreMorphs(InputField field, bool resync = true)
        {
            if (this.MorphBanks != null && MorphFilterPlus.IgnoredMorphs != null)
            {
                if (field == this.NewFieldFemale)
                {
                    this.RestoreBank(this.MorphBanks[MorphFilterPlus.BankIndex.FemaleChar]);
                    this.RestoreBank(this.MorphBanks[MorphFilterPlus.BankIndex.FemaleGen]);
                }
                else if (field == this.NewFieldMale)
                {
                    this.RestoreBank(this.MorphBanks[MorphFilterPlus.BankIndex.MaleChar]);
                    this.RestoreBank(this.MorphBanks[MorphFilterPlus.BankIndex.MaleGen]);
                    this.RestoreBank(this.MorphBanks[MorphFilterPlus.BankIndex.MaleGen2]);
                }
                if (resync) this.ResyncMorphs(field);
            }
        }

        /// <summary>
        /// iterates all morphs in a bank to restore them to their pristine visibility.
        /// </summary>
        private void RestoreBank(DAZMorphBank bank)
        {
            if (bank != null)
            {
                foreach (var morph in bank.morphs)
                {
                    morph.visible = !MorphFilterPlus.IgnoredMorphs.Contains(morph.uid);
                }
            }
        }

        /// <summary>
        /// Syncs the internal morph list of both UIs for the person.
        /// </summary>
        public void ResyncMorphs(bool visibleOnly = false)
        {
            if (this.UIMale != null && (!visibleOnly||this.UIMale.gameObject.activeInHierarchy)) this.UIMale.ResyncMorphs();
            if (this.UIFemale != null && (!visibleOnly || this.UIFemale.gameObject.activeInHierarchy)) this.UIFemale.ResyncMorphs();
        }

        /// <summary>
        /// Syncs the internal morph list of the ui associated with the input field.
        /// </summary>
        /// <param name="field"></param>
        public void ResyncMorphs(InputField field)
        {
            if (field == this.NewFieldFemale && this.UIFemale != null) this.UIFemale.ResyncMorphs();
            else if (field == this.NewFieldMale && this.UIMale != null) this.UIMale.ResyncMorphs();
        }

        /// <summary>
        /// Refreshes the drawn morphs on morph UI panels if they are active (or if forced)
        /// </summary>
        public void RefreshUI(bool visibleOnly = false)
        {            
            if (this.UIMale != null && (visibleOnly || this.UIMale.gameObject.activeInHierarchy)) this.UIMale.ForceCategoryRefresh();
            else if (this.UIFemale != null && (visibleOnly || this.UIFemale.gameObject.activeInHierarchy)) this.UIFemale.ForceCategoryRefresh();
        }

        /// <summary>
        /// refreshes the drawn morphs on the page associated with a search bar,
        /// if the page is active or if forced.
        /// </summary>
        public void RefreshUI(InputField field, bool forceRefresh = false)
        {
            if (field == this.NewFieldFemale && this.UIFemale != null && (forceRefresh || this.UIFemale.gameObject.activeInHierarchy))
            {
                this.UIFemale.ForceCategoryRefresh();
            }
            else if (field == this.NewFieldMale && this.UIMale != null && (forceRefresh || this.UIMale.gameObject.activeInHierarchy))
            {
                this.UIMale.ForceCategoryRefresh();
            }
        }


        /// <summary>
        /// Destroys all generated objects, removes all modifications,
        /// unsubscribes from everything and dumps all references,
        /// in a complete reversion of the upgrade. Called when the person 
        /// or plugin is removed.
        /// </summary>
        public override void Revert()
        {
            base.Revert();
            this.RestoreMorphs();
            this.RefreshUI();
            if (this.OldFieldFemale != null) this.OldFieldFemale.gameObject.SetActive(true);
            if (this.OldFieldMale != null) this.OldFieldMale.gameObject.SetActive(true);

            if (this.NewFieldFemale != null) UnityEngine.Object.Destroy(this.NewFieldFemale.gameObject);
            if (this.NewFieldMale != null) UnityEngine.Object.Destroy(this.NewFieldMale.gameObject);

            // while not strictly necessary, we want to make sure references are released
            // to avoid memory bloat, in case VAM or Unity or a plugin maintains a reference
            // to this object somewhere, somehow.
            this.OldFieldFemale = null;
            this.OldFieldMale = null;

            this.UIMale = null;
            this.UIFemale = null;
            this.MorphBanks = null;

        }
    }

 }
