namespace GossamerVR.MorphFilterPlus
{
    using System;
    using System.Collections.Generic;
    using UnityEngine;
    using UnityEngine.UI;


    public interface IUIElementBridge
    {
        event EventHandler InteractEvent;
        event EventHandler RemovalEvent;
        void Remove();
    }

    /// <summary>
    /// A container representing a duplication of a button to replace
    /// an existing VAM UI button or render on top to intercept a click. 
    /// Has a public click event and can optionally forward the click on to the original button. 
    /// this is used to ensure we restore morphs before performing the normal button action.
    /// </summary>
    public abstract class UIElementBridge<T> : IUIElementBridge where T : Selectable
    {
        public event EventHandler InteractEvent;
        public event EventHandler RemovalEvent;
        public T NewElement { get; private set; }
        public T OldElement { get; private set; }

        protected bool forwardEvents;
        
        /// <summary>
        /// if true, the new button is hidden in front of the old button.
        /// this is tracked so we don't send events if the old button is not interactable.
        /// </summary>
        protected bool isMimic = false;

        protected UIElementBridge() { }

        public UIElementBridge(T oldElement, bool hideOldElement, bool forwardEvents) 
        {
            if (oldElement != null)
            {
                this.forwardEvents = forwardEvents;
                this.OldElement = oldElement;
                this.NewElement = UnityEngine.Object.Instantiate(oldElement, oldElement.transform.parent);
                this.InitNewElement();
                
                // whether or not we hide the old element let's put this one on top.
                this.NewElement.transform.SetSiblingIndex(oldElement.transform.GetSiblingIndex() + 1);

                this.isMimic = !hideOldElement;
                if (this.isMimic)
                {
                    this.SetupMimic();
                }
                else
                {
                    this.OldElement.gameObject.SetActive(false);
                }
            }
        }

        protected abstract void InitNewElement();

        protected abstract void SetupMimic();

        protected abstract void OnRemove();

        protected void Interact()
        {
            this.InteractEvent?.Invoke(this, EventArgs.Empty);
        }
        public void Remove()
        {
            this.RemovalEvent?.Invoke(this, EventArgs.Empty);
            this.OnRemove();
            UnityEngine.Object.Destroy(this.NewElement.gameObject);
            if (!this.isMimic && this.OldElement != null)
            {
                this.OldElement.gameObject.SetActive(true);
            }
            this.OldElement = null;
            this.NewElement = null;
            this.InteractEvent = null;
            this.RemovalEvent = null;
        }
    }
}